@@ -117,6 +117,12 @@ class ProtocolHandler:
 			if prefix == "I4":
 				# IPv4 address without port. A zero address (0.0.0.0) means that the server should replace it with the real IP of the client.
 				user.ip = item[2:len(item)]
+				if isinstance(self.client.events, dict):
+					if 'oniprecvd' in self.client.events:
+						self.client.events['oniprecvd'](self.client.clientinfo.hubaddress())
+				else:
+					if 'oniprecvd' in dir(self.client.events):
+						self.client.events.oniprecvd(self.client.clientinfo.hubaddress(), user.sid, user.ip)
 			#elif prefix == "I6":
 				# IPv6 address without port. A zero address ([0:0:0:0:0:0:0:0]) means that the server should replace it with the real IP of the client.
 			elif prefix == "U4":
@@ -135,7 +141,7 @@ class ProtocolHandler:
 				user.sharedfiles = int(item[2:len(item)])
 			elif prefix == "VE":
 				# Client identification, version (client specific, recommended a short identifier then a float for version number). It is important that hubs don’t discriminate clients based on their VE tag but instead rely on SUP when it comes to which clients should be allowed (for example, “we only want clients that can hash”). VE is there mainly for informative reasons, and can perhaps be used to warn users that they’re using a known buggy or vulnerable client.
-				user.tag = item[2:len(item)]
+				user.tag = self.dcdecode(item[2:len(item)])
 			elif prefix == "US":
 				# Max upload speed bits/sec, integer
 				user.maxuploadspeed = int(item[2:len(item)])
@@ -156,7 +162,7 @@ class ProtocolHandler:
 				user.username = item[2:len(item)]
 			elif prefix == "DE":
 				# Description, string. Valid are all displayable characters (char code >= 32).
-				user.description = item[2:len(item)]
+				user.description = self.dcdecode(item[2:len(item)])
 			elif prefix == "HN":
 				# Hubs where user is a normal user, integer
 				user.hubsnormal = int(item[2:len(item)])
@@ -172,7 +178,8 @@ class ProtocolHandler:
 			elif prefix == "CT":
 				# 1=bot, 2=registered user, 4=operator, 8=super user, 16=hub owner, 32=hub (used when the hub sends an INF about itself). Multiple types are specified by adding the numbers together.
 				ct = int(item[2:len(item)])
-				if ct == 10 or ct == 18 or ct == 6:
+				user.hubstatus = ct
+				if ct >= 5:
 					user.hubsop += 1
 					if isme:
 						sendbinf += " HO" + str(user.hubsop)
@@ -187,7 +194,7 @@ class ProtocolHandler:
 			elif prefix == "OP":
 				# 1 = op
 				if len(item) > 2:
-					if item[2:1] == "1":
+					if item.strip()[2:] == "1":
 						isflag = True
 					user.isoperator = isflag
 					if isflag:
@@ -195,19 +202,19 @@ class ProtocolHandler:
 			elif prefix == "AW":
 				# 1 = Away Other away modes reserved for the future
 				if len(item) > 2:
-					if item[2:1] == "1":
+					if item[2:] == "1":
 						isflag = True
 					user.isaway = isflag
 			elif prefix == "BO":
 				# 1 = Bot
 				if len(item) > 2:
-					if item[2:1] == "1":
+					if item[2:] == "1":
 						isflag = True
 					user.isbot = isflag
 			elif prefix == "HI":
 				# 1 = Hidden
 				if len(item) > 2:
-					if item[2:1] == "1":
+					if item[2:] == "1":
 						isflag = True
 					user.ishidden = isflag
 			elif prefix == "SU":
@@ -216,13 +223,21 @@ class ProtocolHandler:
 		self.client.nicklist.adduser(user)
 		if isme and len(sendbinf) > 11:
 			if self.client.isconnected == False:
-				if 'onconnected' in self.client.events:
-					self.client.events['onconnected'](self.client.clientinfo.hubaddress())
+				if isinstance(self.client.events, dict):
+					if 'onconnected' in self.client.events:
+						self.client.events['onconnected'](self.client.clientinfo.hubaddress())
+				else:
+					if 'onconnected' in dir(self.client.events):
+						self.client.events.onconnected(self.client.clientinfo.hubaddress())
 			self.client.sendrawmessage(sendbinf + "\n")
 			self.client.isconnected = True
 		if newuser and self.client.isconnected:
-			if 'onjoin' in self.client.events:
-				self.client.events['onjoin'](self.client.clientinfo.hubaddress(), user.username)
+			if isinstance(self.client.events, dict):
+				if 'onjoin' in self.client.events:
+					self.client.events['onjoin'](self.client.clientinfo.hubaddress(), user.username)
+			else:
+				if 'onjoin' in dir(self.client.events):
+					self.client.events.onjoin(self.client.clientinfo.hubaddress(), user.username)
 
 	def doemsg(self, parameter):
 		mess = parameter.split(" ")
@@ -234,11 +249,19 @@ class ProtocolHandler:
 		text = self.dcdecode(mess[2])
 		if u.username != "":
 			if len(mess) == 5 and mess[4] == "ME1":
-				if 'onprivateemote' in self.client.events:
-					self.client.events['onprivateemote'](self.client.clientinfo.hubaddress(), u, text)
+				if isinstance(self.client.events, dict):
+					if 'onprivateemote' in self.client.events:
+						self.client.events['onprivateemote'](self.client.clientinfo.hubaddress(), u, text)
+				else:
+					if 'onprivateemote' in dir(self.client.events):
+						self.client.events.onprivateemote(self.client.clientinfo.hubaddress(), u, text)
 			elif len(mess) == 4:
-				if 'onprivatemessage' in self.client.events:
-					self.client.events['onprivatemessage'](self.client.clientinfo.hubaddress(), u, text)
+				if isinstance(self.client.events, dict):
+					if 'onprivatemessage' in self.client.events:
+						self.client.events['onprivatemessage'](self.client.clientinfo.hubaddress(), u, text)
+				else:
+					if 'onprivatemessage' in dir(self.client.events):
+						self.client.events.onprivatemessage(self.client.clientinfo.hubaddress(), u, text)
 
 	def domsg(self, message):
 		mess = message.split(" ")
@@ -246,17 +269,29 @@ class ProtocolHandler:
 		text = self.dcdecode(mess[1])
 		if u.username != "":
 			if len(mess) == 3 and mess[2] == "ME1":
-				if 'onpublicemote' in self.client.events:
-					self.client.events['onpublicemote'](self.client.clientinfo.hubaddress(), u, text)
+				if isinstance(self.client.events, dict):
+					if 'onpublicemote' in self.client.events:
+						self.client.events['onpublicemote'](self.client.clientinfo.hubaddress(), u, text)
+				else:
+					if 'onpublicemote' in dir(self.client.events):
+						self.client.events.onpublicemote(self.client.clientinfo.hubaddress(), u, text)
 			elif len(mess) == 2:
-				if 'onpublicmessage' in self.client.events:
-					self.client.events['onpublicmessage'](self.client.clientinfo.hubaddress(), u, text)
-	
+				if isinstance(self.client.events, dict):
+					if 'onpublicmessage' in self.client.events:
+						self.client.events['onpublicmessage'](self.client.clientinfo.hubaddress(), u, text)
+				else:
+					if 'onpublicmessage' in dir(self.client.events):
+						self.client.events.onpublicmessage(self.client.clientinfo.hubaddress(), u, text)
+
 	def doqui(self, parameter):
 		user = self.client.nicklist.get(parameter)
 		if user:
-			if 'onpart' in self.client.events:
-				self.client.events['onpart'](self.client.clientinfo.hubaddress(), user.username)
+			if isinstance(self.client.events, dict):
+				if 'onpart' in self.client.events:
+					self.client.events['onpart'](self.client.clientinfo.hubaddress(), user.username)
+			else:
+				if 'onpart' in dir(self.client.events):
+					self.client.events.onpart(self.client.clientinfo.hubaddress(), user.username)
 			if user.isoperator:
 				self.client.nicklist.removeoperator(user.sid)
 			self.client.nicklist.removebyclass(user)
@@ -265,8 +300,12 @@ class ProtocolHandler:
 		if self.client.clientinfo.followredirects:
 			self.client.disconnect()
 			self.client.clientinfo.hubaddress(parameter)
-			if 'onredirect' in self.client.events:
-				self.client.events['onredirect'](self.client.clientinfo.hubaddress())
+			if isinstance(self.client.events, dict):
+				if 'onredirect' in self.client.events:
+					self.client.events['onredirect'](self.client.clientinfo.hubaddress())
+			else:
+				if 'onredirect' in dir(self.client.events):
+					self.client.events.onredirect(self.client.clientinfo.hubaddress())
 			self.client.connect()
 	
 	def doquipar(self, par):
@@ -303,8 +342,12 @@ class ProtocolHandler:
 		th = tiger.new(out.getvalue())
 		final = b32encode(th.digest())
 		self.client.sendrawmessage("HPAS " + final[0:final.find("=")])
-		if 'onpassword' in self.client.events:
-			self.client.events['onpassword'](self.client.clientinfo.hubaddress())
+		if isinstance(self.client.events, dict):
+			if 'onpassword' in self.client.events:
+				self.client.events['onpassword'](self.client.clientinfo.hubaddress())
+		else:
+			if 'onpassword' in dir(self.client.events):
+				self.client.events.onpassword(self.client.clientinfo.hubaddress())
 	
 	def dosid(self, parameter):
 		self.client.sid = parameter
@@ -328,17 +371,29 @@ class ProtocolHandler:
 		status = parameter.split(" ")
 		scode = int(status[0])
 		if scode < 200:
-			if 'onstatusmessage' in self.client.events:
-				self.client.events['onstatusmessage'](self.client.clientinfo.hubaddress(), scode, status[1])
+			if isinstance(self.client.events, dict):
+				if 'onstatusmessage' in self.client.events:
+					self.client.events['onstatusmessage'](self.client.clientinfo.hubaddress(), scode, status[1])
+			else:
+				if 'onstatusmessage' in dir(self.client.events):
+					self.client.events.onstatusmessage(self.client.clientinfo.hubaddress(), scode, status[1])
 		elif scode == 200:
 			return
 		elif scode < 300:
 			if scode == 223:
-				if 'onbadpass' in self.client.events:
-					self.client.events['onbadpass'](self.client.clientinfo.hubaddress())
+				if isinstance(self.client.events, dict):
+					if 'onbadpass' in self.client.events:
+						self.client.events['onbadpass'](self.client.clientinfo.hubaddress())
+				else:
+					if 'onbadpass' in dir(self.client.events):
+						self.client.events.onbadpass(self.client.clientinfo.hubaddress())
 			else:
-				if 'onstatusmessage' in self.client.events:
-					self.client.events['onstatusmessage'](self.client.clientinfo.hubaddress(), scode, status[1])
+				if isinstance(self.client.events, dict):
+					if 'onstatusmessage' in self.client.events:
+						self.client.events['onstatusmessage'](self.client.clientinfo.hubaddress(), scode, status[1])
+				else:
+					if 'onstatusmessage' in dir(self.client.events):
+						self.client.events.onstatusmessage(self.client.clientinfo.hubaddress(), scode, status[1])
 				if self.client.clientinfo.do_ssl:
 					self.client.sslsock.disconnect()
 				else:
@@ -356,12 +411,20 @@ class ProtocolHandler:
 			if p[0:2] == "NI": # Hub name
 				self.client.hubinfo.hubname = self.dcdecode(p[2:len(p)])
 				if self.dcdecode(p[2:len(p)]) != " ":
-					if 'onhubname' in self.client.events:
-						self.client.events['onhubname'](self.client.clientinfo.hubaddress(), self.client.hubinfo.hubname)
+					if isinstance(self.client.events, dict):
+						if 'onhubname' in self.client.events:
+							self.client.events['onhubname'](self.client.clientinfo.hubaddress(), self.client.hubinfo.hubname)
+					else:
+						if 'onhubname' in dir(self.client.events):
+							self.client.events.onhubname(self.client.clientinfo.hubaddress(), self.client.hubinfo.hubname)
 			elif p[0:2] == "DE": # Topic
 				self.client.hubinfo.topic = self.dcdecode(p[2:len(p)])
-				if 'ontopic' in self.client.events:
-					self.client.events['ontopic'](self.client.clientinfo.hubaddress(), self.client.hubinfo.topic)
+				if isinstance(self.client.events, dict):
+					if 'ontopic' in self.client.events:
+						self.client.events['ontopic'](self.client.clientinfo.hubaddress(), self.client.hubinfo.topic)
+				else:
+					if 'ontopic' in dir(self.client.events):
+						self.client.events.ontopic(self.client.clientinfo.hubaddress(), self.client.hubinfo.topic)
 			elif p[0:2] == "VE": # Hub Version
 				self.client.hubinfo.hubversion = self.dcdecode(p[2:len(p)])
 			#elif p[0:2] == "RE": # Redirect
